/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

#include "Midiator.h"
#include "err_and_warn.h"
#include <stdio.h> // for sprintf
#include <process.h> // for _beginthreadex

void Midiator::open_if_needed()
{
	if (!initialized)
		comSetup();
	initialized = True;
}

void Midiator::close_if_needed()
{
	if (open == False && output_open[0] == False && output_open[1] == False &&
		output_open[2] == False && output_open[3] == False)
	{
		if (initialized && !CloseHandle(hCom))
			ew.w("Unable to close %s", portname);
		initialized = False;
	}
}

Midiator::Midiator(const char* comport, void (*func)(HMIDIIN, UINT, DWORD, DWORD, DWORD))
{
	strncpy(portname, comport, 31);
	callback = func;
	open = False;
	output_open[0] = False;
	output_open[1] = False;
	output_open[2] = False;
	output_open[3] = False;
	initialized = False;
}

Midiator::~Midiator()
{
	closeInput(0);
	while (open)
		Sleep(100);
	close_if_needed();
}

void Midiator::getInputName(UINT n, char* s)
{
	sprintf(s,"MIDIator MS-124W input on %s",portname);
}

void Midiator::getOutputName(UINT n, char* s)
{
	sprintf(s,"MIDIator MS-124W output %u on %s",n+1,portname);
}

void Midiator::openInput(UINT n)
{
	open_if_needed();
	if (!open)
	{
		stopflag = False;
		open = True;
		threadHandle = (HANDLE) _beginthreadex(NULL, 0, threadMain, (void*)this, 0, &threadID);
		SetThreadPriority(threadHandle, THREAD_PRIORITY_LOWEST);
	}
}

void Midiator::closeInput(UINT n)
{
	if (open)
		stopflag = True;
	close_if_needed();
}

void Midiator::openOutput(UINT n)
{
	open_if_needed();
	output_open[n] = True;
	running_status = 0;
	outport = -1;
}

void Midiator::closeOutput(UINT n)
{
	output_open[n] = False;
	close_if_needed();
}

void Midiator::setRts(int x)
{
	if (!EscapeCommFunction(hCom, (x ? SETRTS : CLRRTS)))
		ew.w("Failed to %s RTS on %s", (x ? "set" : "clear"), portname);
}

void Midiator::setDtr(int x)
{
	if (!EscapeCommFunction(hCom, (x ? SETDTR : CLRDTR)))
		ew.w("Failed to %s DTR on %s", (x ? "set" : "clear"), portname);
}

void Midiator::send(UINT n, DWORD p)
{
	const char     msg_len[ 7] = {	3, 3, 3, 3, 2, 2, 3 };//8	9	A	B	C	D	E	F
	const char sys_msg_len[16] = {	0, 2, 3, 2, 0, 0, 1, 1,	1, 0, 1, 1, 1, 0, 1, 1 };

	unsigned char status = (unsigned char)(p & 0xff);
	unsigned char d1 = (unsigned char)((p >> 8) & 0xff);
	unsigned char d2 = (unsigned char)((p >> 16) & 0xff);

//	DWORD foo;

	if (((int)n) != outport) // we need to change port
	{
		running_status = 0; // invalidate running status
		outport = n; // set new output port
		setRts(1); // set status bit = 1
		setDtr(0); // bring clock down
		Sleep(0); // give up our timeslice - what we really
		          // want to do is delay 4 microseconds
		setDtr(1); // rising edge latches data
		Sleep(0);
		setDtr(0);
		setRts(outport >> 1); // msb of port
		Sleep(0);
		setDtr(1); // rising edge
		Sleep(0);
		setDtr(0);
		setRts(outport & 1); // lsb of port
		Sleep(0);
		setDtr(1); // rising edge
		Sleep(0);
	}

	if (status < 128)
	{
		ew.w("status byte has hi bit clear -- not a valid MIDI message");
		return;
	}

//	tw.pr("sending status byte 0x%x\r", status);
	if (status != running_status || (status >> 4 == system))
	{
		if (!TransmitCommChar(hCom, status))
//		if (!WriteFile(hCom, &status, 1, &foo, NULL))
			ew.w("unable to write to %s", portname);
		running_status = status;
		if (status >> 4 == system && sys_msg_len[status & 0xf] < 2)
			return;
	}

	if (d1 > 127)
	{
		ew.w("data byte 1 has hi bit set");
		d1 &= 0x7f;
	}

//	tw.pr("sending data byte 1 0x%x\r", d1);
	if (!TransmitCommChar(hCom, d1))
//	if (!WriteFile(hCom, &d1, 1, &foo, NULL))
		ew.w("unable to write to %s", portname);
	if ((((status & 0x7f) >> 4) == system && sys_msg_len[status & 0xf] == 2) ||
		 (msg_len[(status & 0x7f) >> 4] == 2))
		 return;

	if (d2 > 127)
	{
		ew.w("data byte 2 has hi bit set");
		d2 &= 0x7f;
	}

//	tw.pr("sending data byte 2 0x%x\r", d2);
	if (!TransmitCommChar(hCom, d2))
//	if (!WriteFile(hCom, &d2, 1, &foo, NULL))
		ew.w("unable to write to %s", portname);
}

unsigned WINAPI Midiator::threadMain(LPVOID lpvThreadParm)
{
	Midiator* that = (Midiator*) lpvThreadParm;
	that->inputloop();
	that->open = False;
	return 0;
}

void Midiator::comSetup()
{
	DCB dcb;
	DWORD dwError;
	BOOL fSuccess;
	
	hCom = CreateFile(portname,
		GENERIC_READ | GENERIC_WRITE,
		0,    /* comm devices must be opened w/exclusive-access */
		NULL, /* no security attrs */
		OPEN_EXISTING, /* comm devices must use OPEN_EXISTING */
		0,    /* not overlapped I/O */
		NULL  /* hTemplate must be NULL for comm devices */
		);
	
	if (hCom == INVALID_HANDLE_VALUE) {
		dwError = GetLastError();
	
		/* handle error */
		ew.w("Error opening %s",portname);
		return;
	}
	
	/*
	* Omit the call to SetupComm to use the default queue sizes.
	* Get the current configuration.
	*/
	
	fSuccess = GetCommState(hCom, &dcb);
	
	if (!fSuccess) {
		/* Handle the error. */
		ew.w("Error getting %s state",portname);
		return;
	}
	
	/* Fill in the DCB: baud=38400, 8 data bits, no parity, 1 stop bit. */

//	tw.pr("foo bar\r");
	
	dcb.BaudRate = 38400;
	dcb.fBinary = TRUE;
	dcb.fParity = FALSE;
	dcb.fOutxCtsFlow = TRUE;
	dcb.fDtrControl = DTR_CONTROL_ENABLE;
	dcb.fOutX = FALSE;
	dcb.fInX = FALSE;
	dcb.fNull = FALSE;
	dcb.fRtsControl = RTS_CONTROL_ENABLE;
	dcb.ByteSize = 8;
	dcb.Parity = NOPARITY;
	dcb.StopBits = ONESTOPBIT;
	
	fSuccess = SetCommState(hCom, &dcb);
	
	if (!fSuccess) {
	    /* Handle the error. */
		ew.w("Error setting %s state",portname);
		return;
	}

	COMMTIMEOUTS cto;
	cto.ReadIntervalTimeout = MAXDWORD;
	cto.ReadTotalTimeoutMultiplier = 0;
	cto.ReadTotalTimeoutConstant = 0;
	cto.WriteTotalTimeoutMultiplier = 0;
	cto.WriteTotalTimeoutConstant = 500;
	if (SetCommTimeouts(hCom, &cto) == FALSE)
		ew.w("Error setting %s timeouts",portname);
}

void Midiator::inputloop()
{
	const char     msg_len[ 7] = {	3, 3, 3, 3, 2, 2, 3 };//8	9	A	B	C	D	E	F
	const char sys_msg_len[16] = {	0, 2, 3, 2, 0, 0, 1, 1,	1, 0, 1, 1, 1, 0, 1, 1 };

	unsigned char status = 0;
	unsigned char buf[128];
	while (TRUE)
	{
		DWORD bytesread;
		unsigned char d1;
		if (ReadFile(hCom, buf, 1, &bytesread, NULL) == FALSE)
		{
			ew.w("Error occurred reading byte from %s",portname);
			continue;
		}
		d1 = buf[0];
		if (bytesread == 0) // no byte waiting
		{
			if (stopflag)
				return;
			else
				continue;
		}
//		tw.pr("%02x ",d1);
//		DWORD modemstatus;
//		if (GetCommModemStatus(hCom, &modemstatus) == FALSE)
//		{
//			ew.w("Error occurred calling GetCommModemStatus on %s",portname);
//			continue;
//		}
//		tw.spr("We read 0x%x\n",d1);
		if (d1 & 0x80) // hi bit on - indicates MIDI status byte
		{
//			tw.spr("We MIDI status\r");
			status = d1 | 0x80;
			if (d1 >> 4 == system && sys_msg_len[d1 & 0xf] < 2)
			{
				callback(NULL, MIM_DATA, 0, (DWORD)status, in_start);
				status = 0;
			}
			continue;
		}
		if (status == 0)
		{
			ew.w("A data byte was received before any status on %s",portname);
			continue;
		}
		if ((((status & 0x7f) >> 4) == system && sys_msg_len[status & 0xf] == 2) ||
			 (msg_len[(status & 0x7f) >> 4] == 2))
		{
			callback(NULL, MIM_DATA, 0, ((DWORD)status) | (((DWORD)d1) << 8), in_start);
			continue;
		}
		unsigned char d2;
		do
		{
			if (ReadFile(hCom, buf, 1, &bytesread, NULL) == FALSE)
			{
				ew.w("Error occurred reading byte from %s",portname);
				continue;
			}
		} while (bytesread == 0);
		d2 = buf[0];
//		tw.pr("%02x ",d2);
//		if (bytesread == 0) // we time out for a half second
//		{
//			ew.w("Timed out waiting for second data byte on %s", portname);
//			continue;
//		}
//		tw.spr("We read 0x%x\n",d2);
		callback(NULL, MIM_DATA, 0, ((DWORD)status) | (((DWORD)d1) << 8) | (((DWORD)d2) << 16), in_start);
	}
}

// Author: Patrick Pelletier
