/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

#include "Score.h"
#include "err_and_warn.h"
#include "TrackMerger.h"
#include "OMS_app.h"
#include "Recorder.h"
#include "ScoreCreationInfo.h"

void Score::warnfunc(const char *s)
{
	ew.w("%s", s);
}

void Score::convertFuncTime(ScoreTimeArrayPtr stap, ScoreTime lastTime, long lastTick, long thisTick, int* func_index, long currentTempo, long division)
{
	int i;
	for (i = *func_index; i < stap->length; i++)
	{
		if (thisTick != -1 && (*stap)[i] > thisTick)
		{
			*func_index = i;
			return;
		}
		long old = (*stap)[i];
		double multiplier = double(currentTempo)/double(division)/double(500);
		(*stap)[i] = (ScoreTime) (lastTime + ((*stap)[i]-lastTick)*multiplier + 0.5);
	}
	*func_index = i;
}

void Score::convertTime(MIDI_msgArrayPtr p, long division, ScoreTimeArrayPtr stap)
{
	if (p->length < 1)
		return;
	// change tick time into real time
	int func_index = 0;
	ScoreTime lastTime = 0;
	long lastTick = 0;
	long currentTempo = 500000;

	for (int i = 0; i < p->length; i++)
	{
		MIDI_msg* m = &((*p)[i]);
		long thisTick = m->time;
		double multiplier = double(currentTempo)/double(division)/double(500);
		
		m->time = (long) (lastTime + (thisTick-lastTick)*multiplier + 0.5);
		
		if (m->is_ext() && m->ed->st == smsgt::file_meta &&
			 m->ed->meta_event.mt == meta::tempo)
		{
			convertFuncTime(stap, lastTime, lastTick, thisTick, &func_index, currentTempo, division);
			long newTempo = m->ed->meta_event.tem;
			lastTime += (ScoreTime) ((thisTick-lastTick)*multiplier + 0.5);
			lastTick = thisTick;
			currentTempo = newTempo;
		}
	}
	convertFuncTime(stap, lastTime, lastTick, -1, &func_index, currentTempo, division);
}

void Score::read(const Filedata* f)
{
	int2 numtracks = f->getNumTracks();
	int2 division = f->getDivision();
	if (numtracks < 1)
	{
		msgs = new MIDI_msgArray(0);
		return;
	}
	ioref node = oap->open_output();
	MIDI_msgArrayPtrArrayPtr p = new MIDI_msgArrayPtrArray(numtracks);
	for (int i = 0; i < numtracks; i++)
	{
		Filedata::MidiMsgArrayPtr p1 = f->getTrack(i);
		MIDI_msgArrayPtr p2 = new MIDI_msgArray(p1->length);
		if (p1->length < 1)
			continue;
		int j;
		for (j = 0; j < p1->length; j++) // copy MidiMsgs to MIDI_msgs
		{
			MidiMsg* m = &((*p1)[j]);
			m->sanityCheck();
			*((MidiMsg*)&(*p2)[j]) = *m;
			(*p2)[j].node = node;
		}
		(*p)[i] = p2;
	}
	msgs = TrackMerger<MIDI_msg>::merge(p); // merge all the tracks together
	funcs = new ScoreFuncArray(0);
	args = new VoidStarArray(0);
	times = new ScoreTimeArray(0);
	convertTime(msgs, division, times);
}

Score::Score(const ScoreCreationInfo* sci_const)
{
	ScoreCreationInfo* sci = (ScoreCreationInfo*) sci_const; // cast away constness because we maintain logical constness
	sci->do_sort();
	int i;
	int2 numtracks = 0;
	for (i = 0; i < sci->track_included.length; i++)
		if (sci->track_included[i])
			numtracks++;
	if (sci->num_msgs > 0)
		numtracks++;
	int2 division = sci->division;
	if (numtracks < 1)
	{
		msgs = new MIDI_msgArray(0);
		return;
	}
	MIDI_msgArrayPtrArrayPtr p = new MIDI_msgArrayPtrArray(numtracks);
	int i2 = 0;
	for (i = 0; i < sci->track_included.length; i++)
	{
		if (!sci->track_included[i])
			continue;
		ioref node = sci->track_node[i];
		Filedata::MidiMsgArrayPtr p1 = sci->fd->getTrack(i);
		MIDI_msgArrayPtr p2 = new MIDI_msgArray(p1->length);
		if (p1->length < 1)
			continue;
		int j;
		for (j = 0; j < p1->length; j++) // copy MidiMsgs to MIDI_msgs
		{
			MidiMsg* m = &((*p1)[j]);
			m->sanityCheck();
			*((MidiMsg*)&(*p2)[j]) = *m;
			(*p2)[j].node = node;
		}
		(*p)[i2++] = p2;
	}
	if (sci->num_msgs > 0) // handle manually inserted MIDI_msgs
	{
		MIDI_msgArrayPtr p2 = new MIDI_msgArray(sci->num_msgs);
		int j;
		for (j = 0; j < p2->length; j++)
			(*p2)[j] = *(sci->msgs[j]);
		(*p)[i2++] = p2;
	}
	if (i2 != numtracks)
		ew.w("Sanity check failure in Score::Score(const ScoreCreationInfo*)");
	msgs = TrackMerger<MIDI_msg>::merge(p); // merge all the tracks together
	funcs = new ScoreFuncArray(sci->num_funcs);
	args = new VoidStarArray(sci->num_funcs);
	times = new ScoreTimeArray(sci->num_funcs);
	int j;
	for (j = 0; j < sci->num_funcs; j++)
	{
		(*funcs)[j] = sci->funcs[j]->func;
		(*args)[j] = sci->funcs[j]->v;
		(*times)[j] = sci->funcs[j]->time;
	}
	convertTime(msgs, division, times);
}


ScoreTime Score::getLength() const
{
	if (msgs->length < 1)
		return 0;
	return (*msgs)[msgs->length-1].time; // return time of last event
}

static void iterator(ioref, int, Bool* e, void*)
{
	*e = 0;
}

typedef PerChannel<Bool> ChannelUseMap;
typedef UtilPtr<ChannelUseMap> ChannelUseMapPtr;

ChannelUseMapPtr Score::getChannels() const
{
	if (u != NULL)
		return u;
	ChannelUseMapPtr result = new ChannelUseMap;
	result->iterate(iterator, NULL);  // initialize to zero
	if (msgs->length)
		for (int i = 0; i < msgs->length; i++)
		{
			MIDI_msg* m = &((*msgs)[i]);
			m->sanityCheck();
			if (m->msgt() != MIDI_msg::system)
				(*result)(m->node, m->chan()) = 1;
		}
	((Score*)this)->u = result; // cache result: logical, not physical constness
	return result;
}

Score::Score(const Recorder* r)
{
	msgs = new MIDI_msgArray(r->current_index);
	if (r->current_index == 0)
		return;
	Recorder::storage_group* p = r->first;
	int offset = 0;
	long mintime = (*p)[0].time;
	for (int i = 0; i < msgs->length; i++) // copy contents of Recorder
	{
		(*msgs)[i] = (*p)[i-offset];
		(*msgs)[i].time -= mintime; // adjust time
		if (i-offset > Recorder::storage_group::group_size)
		{
			offset += Recorder::storage_group::group_size;
			p = p->next;
		}
	}
	funcs = new ScoreFuncArray(0);
	args = new VoidStarArray(0);
	times = new ScoreTimeArray(0);
}

// Author: Patrick Pelletier
